import os
import ConfigParser
from rptk.base import BaseObject
from rptk.load import ClassLoader


class Dispatcher(BaseObject):
    def __init__(self, **kwargs):
        super(Dispatcher, self).__init__()
        self.log_init()
        self._opts = dict()
        default_config_file = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'rptk.conf')
        self._config_file = kwargs.pop("config_file", default_config_file)
        self.log.debug(msg="reading config file")
        reader = self._read_config()
        self.log.debug(msg="getting default options")
        opts = dict(reader.items("defaults"))
        self.log.debug(msg="updating options with user supplied values")
        opts.update(kwargs)
        self.update(**opts)
        self.log.debug(msg="getting dynamic class loaders")
        try:
            self._query_class_loader = ClassLoader(
                items=reader.items(section="query-classes")
            )
            self._format_class_loader = ClassLoader(
                items=reader.items(section="format-classes")
            )
        except Exception as e:
            self.log.error(msg=e.message)
            raise e
        self.log_init_done()

    def _read_config(self):
        self.log_method_enter(method=self.current_method)
        reader = ConfigParser.SafeConfigParser()
        self.log.debug(msg="trying to read configuration from file %s" % self.config_file)
        try:
            reader.read(self.config_file)
        except Exception as e:
            self.log.error(msg=e.message)
            raise e
        self.log_method_exit(method=self.current_method)
        return reader

    def update(self, **opts):
        self.log_method_enter(method=self.current_method)
        self._opts.update(opts)
        for key in self.opts:
            try:
                if key[0] == '_':
                    raise ValueError("illegal option key: %s" % key)
                val = self.opts[key]
                if val:
                    self.log.debug(msg="updating option: %s = %s" % (key, val))
                    setattr(self, key, val)
            except Exception as e:
                self.log.error(msg=e.message)
                raise e
        self.log_method_exit(method=self.current_method)
        return self

    @property
    def config_file(self):
        return self._config_file

    @property
    def query_class_loader(self):
        return self._query_class_loader

    @property
    def format_class_loader(self):
        return self._format_class_loader

    @property
    def query_class(self):
        try:
            return self.query_class_loader.get_class(
                name=self.query
            )
        except Exception as e:
            self.log.error(msg=e.message)
            raise e

    @property
    def format_class(self):
        try:
            return self.format_class_loader.get_class(
                name=self.format
            )
        except Exception as e:
            self.log.error(msg=e.message)
            raise e

    def dispatch(self, obj=None, name=None, test=False):
        self.log_method_enter(method=self.current_method)
        if not name:
            self.log.debug(msg="name not provided using object (%s)" % obj)
            if obj:
                name = obj
            else:
                name = self.opts["object"]
        self.log.debug(msg="trying to begin query")
        try:
            with self.query_class(**self.opts) as q:
                result = q.query(obj=obj)
        except Exception as e:
            self.log.error(msg=e.message)
            raise e
        self.log.debug(msg="trying to format result for output")
        try:
            with self.format_class(**self.opts) as f:
                output = f.format(result=result, name=name)
                if test:
                    return f.validate(output=output)
        except Exception as e:
            self.log.error(msg=e.message)
            raise e
        self.log_method_exit(method=self.current_method)
        return output



import jinja2
from rptk.base import BaseObject


class BaseFormat(BaseObject):
    def __init__(self, **opts):
        super(BaseFormat, self).__init__()
        self.log_init()
        self._opts = opts
        self.log_init_done()

    def format(self, result=None, name=None):
        self.log_method_enter(method=self.current_method)
        if not isinstance(result, dict):
            self.raise_type_error(arg=result, cls=dict)
        if not name:
            self.log.debug(msg="using name from configuration")
            name = self.name
        if not isinstance(name, basestring):
            self.raise_type_error(arg=name, cls=basestring)
        output = unicode(name)
        self.log_method_exit(method=self.current_method)
        return output

    def validate(self, output=None):
        self.log_method_enter(method=self.current_method)
        if not isinstance(output, basestring):
            self.raise_type_error(arg=output, cls=basestring)
        self.log.debug(msg="validation successful")
        self.log_method_exit(method=self.current_method)
        return True

    @property
    def name(self):
        return self.opts["name"]


class JinjaFormat(BaseFormat):
    template_name = None

    def __init__(self, **opts):
        super(JinjaFormat, self).__init__(**opts)
        self.log.debug("configuring jinja2 environment")
        try:
            self.env = jinja2.Environment(
                loader=jinja2.PackageLoader('rptk')
            )
            self.env.trim_blocks = True
            self.env.lstrip_blocks = True
        except Exception as e:
            self.raise_runtime_error(e.message)
        self._template = None
        self.log_init_done()

    def __enter__(self):
        self.log_ready_start()
        self._load_template()
        self.log_ready_done()
        return self

    @property
    def template(self):
        return self._template

    def _load_template(self):
        try:
            self._template = self.env.get_template(self.template_name)
        except jinja2.TemplateError as e:
            self.log.error(msg=e.message)
            raise
        self.log.debug("template loaded successfully")

    def format(self, result=None, name=None):
        self.log_method_enter(method=self.current_method)
        name = super(JinjaFormat, self).format(result=result, name=name)
        if isinstance(self.template, jinja2.Template):
            try:
                output = self.template.render(result=result, name=name)
                self.log_method_exit(method=self.current_method)
                return output
            except Exception as e:
                self.log.error(msg=e.message)
                raise
        else:
            self.raise_type_error(arg=self.template, cls=jinja2.Template)



import sys
import logging
import argparse
from rptk import dispatch


def main():
    # setup logger
    log = logging.getLogger(__name__)
    lh = logging.StreamHandler()
    lf = logging.Formatter(fmt="%(asctime)s %(name)s: %(levelname)s %(message)s")
    lh.setFormatter(lf)
    logging.getLogger().addHandler(lh)
    # get config_file and debug options
    try:
        parser = argparse.ArgumentParser(add_help=False, argument_default=argparse.SUPPRESS)
        parser.add_argument('--debug', '-d', action='store_true',help="print debug logging output")
        parser.add_argument('--config_file', '-f', type=str, help="path to configuration file")
        args, args_remaining = parser.parse_known_args()
    except Exception as e:
        log.error(msg=e.message)
        return 1
    # set log level
    try:
        if args.debug:
            logging.getLogger().setLevel(logging.DEBUG)
            log.debug(msg="debug logging started")
    except AttributeError:
        pass
    except Exception as e:
        log.error(msg=e.message)
        return 1
    # set up dispatcher with default options
    try:
        log.debug(msg="creating dispatcher object")
        dispatcher = dispatch.Dispatcher(**vars(args))
        log.debug(msg="dispatcher ready")
    except Exception as e:
        log.error(msg=e.message)
        return 1
    # parse remaining args and update dispatcher options
    try:
        parser.add_argument(
            '--query', '-Q', help="query class",
            choices=dispatcher.query_class_loader.class_names
        )
        parser.add_argument(
            '--format', '-F', help="format class",
            choices=dispatcher.format_class_loader.class_names
        )
        parser.add_argument('--host', '-h', type=str, help="irrd host to connect to")
        parser.add_argument('--port', '-p', type=int, help="irrd service tcp port")
        parser.add_argument('--name', '-n', type=str, help="prefix-list name (default: object)")
        parser.add_argument('--help', action='help', help="print usage information and exit")
        parser.add_argument('object', type=str, help="rpsl object name")
        log.debug(msg="parsing command-line arguments")
        args = parser.parse_args(args=args_remaining)
        log.debug(msg="updating dispatcher options")
        dispatcher.update(**vars(args))
    except Exception as e:
        log.error(msg=e.message)
        return 1
    # dispatch query and print formatted result
    try:
        log.debug(msg="dispatching query")
        result = dispatcher.dispatch()
        log.debug(msg="got result length %s" % len(result))
        sys.stdout.write("%s\n" % result)
        return 0
    except Exception as e:
        log.error(msg=e.message)
        return 1

if __name__ == "__main__":
    exit_code = main()
    exit(exit_code)



from flask import Flask, request, make_response
from rptk import RptkAPI

app = Flask(__name__)


@app.route("/formats")
def get_formats():
    opts = request.args.to_dict()
    rptk = RptkAPI(**opts)
    formats = rptk.available_formats()
    response = make_response("\n".join(formats))
    response.headers['Content-Type'] = "text/plain"
    return response


@app.route("/<string:fmt>/<string:obj>")
def get_prefix_list(fmt=None, obj=None):
    opts = request.args.to_dict()
    rptk = RptkAPI(object=obj, format=fmt, **opts)
    result = rptk.query()
    response = make_response(result)
    response.headers['Content-Type'] = "text/plain"
    return response

if __name__ == "__main__":
    app.run(host='::', port=8080)



import yaml
from rptk.modules.format import BaseFormat


class YamlFormat(BaseFormat):
    def format(self, result=None, name=None):
        self.log_method_enter(method=self.current_method)
        name = super(YamlFormat, self).format(result=result, name=name)
        self.log.debug(msg="creating json output")
        try:
            output = yaml.dump(
                {name: result},
                indent=4,
                explicit_start=True,
                explicit_end=True,
                default_flow_style=False,
            )
            self.log_method_exit(method=self.current_method)
            return output
        except Exception as e:
            self.log.error(msg=e.message)
            raise

    def validate(self, output=None):
        self.log_method_enter(method=self.current_method)
        super(YamlFormat, self).validate(output=output)
        try:
            yaml.load(output)
        except Exception as e:
            self.log.error(msg=e.message)
            raise
        self.log_method_exit(method=self.current_method)
        return True



from rptk import dispatch


class Rptk(object):
    def __init__(self, **opts):
        self._dispatcher = dispatch.Dispatcher(**opts)

    def query(self, obj=None, name=None, test=False):
        return self._dispatcher.dispatch(obj=obj, name=name, test=test)

    def available_formats(self):
        return self._dispatcher.format_class_loader.class_names

    def update_opts(self, **opts):
        try:
            self._dispatcher.update(**opts)
        except Exception as e:
            raise e
        return self

    @property
    def query_class_loader(self):
        return self._dispatcher.query_class_loader

    @property
    def format_class_loader(self):
        return self._dispatcher.format_class_loader

    @property
    def opts(self):
        return self._dispatcher.opts

    def __getattribute__(self, name):
        try:
            return super(Rptk, self).__getattribute__(name)
        except AttributeError as e:
            try:
                return self.opts[name]
            except KeyError:
                pass
            raise e

    def __setattr__(self, name, value):
        try:
            self._dispatcher.update(**{name: value})
        except Exception as e:
            super(Rptk, self).__setattr__(name, value)



import os
from rptk import RptkAPI
from unittest import TestCase


class TestRptk(TestCase):
    config_file = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'tests.conf')
    if os.name == 'posix':
        posix = True
    else:
        posix = False

    def test_api(self):
        rptk = RptkAPI(config_file=self.config_file)
        for q in rptk.query_class_loader.class_names:
            query_class = rptk.query_class_loader.get_class(name=q)
            if query_class.posix_only and not self.posix:
                continue
            for f in rptk.format_class_loader.class_names:
                opts = {'query': q, 'format': f}
                rptk.update_opts(**opts)
                result = rptk.query(test=True)
                self.assertTrue(result)
                if result:
                    print "api query with query=%s format=%s passed" % (q, f)



from rptk.base import BaseObject


class BaseQuery(BaseObject):
    posix_only = False

    def __init__(self, **opts):
        super(BaseQuery, self).__init__()
        self.log_init()
        self._opts = opts
        self.log_init_done()

    def query(self, obj=None):
        self.log_method_enter(method=self.current_method)
        if not obj:
            self.log.debug(msg="using object from configuration")
            obj = self.opts["object"]
        if not isinstance(obj, basestring):
            self.raise_type_error(arg=obj, cls=basestring)
        return unicode(obj)

    @property
    def host(self):
        return self.opts["host"]

    @property
    def port(self):
        return int(self.opts["port"])

    @property
    def target(self):
        return "%s:%s" % (self.host, self.port)



import socket
import re
import ipaddress
from pkg_resources import get_distribution
from rptk.modules.query import BaseQuery


class NativeQuery(BaseQuery):
    def __init__(self, **opts):
        super(NativeQuery, self).__init__(**opts)
        self._regexp = re.compile('(?P<state>[ACDEF])(?P<len>\d*)(?P<msg>[\w\s]*)$')
        self._keepalive = True
        self.log_init_done()

    def __enter__(self):
        self.log_ready_start()
        self._connect()
        self.log_ready_done()
        return self

    def __exit__(self, typ, value, traceback):
        self.log_exit_start()
        self._disconnect()
        self.log_exit_done()

    def query(self, obj=None):
        obj = super(NativeQuery, self).query(obj=obj)
        result = dict()
        sets = {u'ipv4': set(), u'ipv6': set()}
        self.log.debug(msg="trying to get members of %s" % obj)
        members = self._members(obj=obj)
        for member in members:
            self.log.debug(msg="trying to get routes for %s" % member)
            routes = self._routes(obj=member)
            for af in routes:
                sets[af].update(routes[af])
        for af in sets:
            prefixes = sorted(list(sets[af]))
            result[af] = [{u'prefix': p.with_prefixlen, u'exact': True} for p in prefixes]
            self.log.debug(msg="found %s %s prefixes for object %s" % (len(result[af]), af, obj))
        self.log_method_exit(method=self.current_method)
        return result

    def _connect(self):
        self.log.debug(msg="creating socket")
        self._socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.log.debug(msg="trying to connect to %s" % self.target)
        try:
            self._socket.connect((self.host, self.port))
        except socket.error as e:
            self.log.error(msg=e.message)
            raise
        self.log.debug(msg="socket connected")
        if self._keepalive:
            self._socket.send('!!\n')
        self._query('!nRPTK-%s' % get_distribution('rptk').version)

    def _disconnect(self):
        self.log.debug(msg="disconnecting socket")
        self._socket.shutdown(socket.SHUT_RDWR)
        self._socket.close()
        self.log.debug(msg="socket closed")

    def _query(self, q, skip_errors=None):
        q += '\n'
        total_sent = 0
        query_length = len(q)
        while total_sent < query_length:
            sent = self._socket.send(q[total_sent:])
            if not sent:
                self.raise_runtime_error(msg="socket connection broken")
            total_sent += sent
        self.log.debug("sent query %s (length %d bytes)", q.rstrip(), total_sent)
        chunks = []
        chunk_size = 4096
        chunk = self._socket.recv(chunk_size)
        response, chunk = chunk.split('\n', 1)
        try:
            response_length = self._parse_response(response)
        except IRRQueryError as e:
            if type(e) in skip_errors:
                self.log.debug(msg=e.message)
                response_length = False
            else:
                self.log.error(msg=e.message)
                raise
        if not response_length:
            return
        total_rcvd = len(chunk) or 0
        chunks.append(chunk)
        while total_rcvd <= response_length:
            self.log.debug("received %d of %d bytes", total_rcvd, response_length)
            chunk = self._socket.recv(chunk_size)
            if chunk == '':
                self.raise_runtime_error(msg="socket connection broken")
            chunks.append(chunk)
            total_rcvd += len(chunk)
        self.log.debug("received %d of %d bytes", total_rcvd, response_length)
        suffix = chunks[-1][-(total_rcvd - response_length):]
        chunks[-1] = chunks[-1][:-len(suffix)]
        self.log.debug("suffix length: %d", len(suffix))
        return ''.join(chunks)

    def _parse_response(self, response):
        """ check response code and return response data length """
        self.log.debug("received response %s", response)
        match = self._regexp.match(response)
        if not match:
            self.raise_runtime_error("invalid response '%s'" % (response,))
        state = match.group('state')
        if state == 'A':
            length = int(match.group('len'))
            self.log.debug(msg="query successful: %d bytes of data" % length)
            return length
        elif state == 'C':
            self.log.debug(msg="query successful. no data.")
            return False
        elif state == 'D':
            raise KeyNotFoundError()
        elif state == 'E':
            raise KeyNotUniqueError()
        elif state == 'F':
            if match.group('msg'):
                msg = match.group('msg').strip()
            else:
                msg = 'unknown error'
            raise OtherError(msg)
        raise RuntimeError("invalid response '%s'" % (response,))

    def _members(self, obj=None):
        """ resolve an as-set to its members """
        q = "!i%s,1" % obj
        members = self._query(q, skip_errors=(KeyNotFoundError,))
        if members:
            members = members.split()
            self.log.debug("found %d members of %s" % (len(members), obj))
            return members
        else:
            self.log.debug("no members of %s found. treating as autnum." % obj)
            return [obj]

    def _routes(self, obj=None):
        """ get routes for specified object """
        proto = {
            u'ipv4': {'cmd': '!g', 'class': ipaddress.IPv4Network},
            u'ipv6': {'cmd': '!6', 'class': ipaddress.IPv6Network}
        }
        routes = dict()
        for af in proto:
            cmd = proto[af]['cmd']
            cls = proto[af]['class']
            q = "%s%s" % (cmd, obj)
            routes[af] = list()
            resp = self._query(q, skip_errors=(KeyNotFoundError,))
            if resp:
                for each in resp.split():
                    try:
                        routes[af].append(cls(unicode(each)))
                    except ipaddress.AddressValueError, ipaddress.NetmaskValueError:
                        self.log.warning("converting %s to %s failed" % (each, cls))
            self.log.debug("found %d %s prefixes for object %s" % (len(routes[af]), af, obj))
        return routes


class IRRQueryError(RuntimeError):
    proto_msg = ''

    def __init__(self, *args, **kwargs):
        super(IRRQueryError, self).__init__(self.proto_msg, *args, **kwargs)


class KeyNotFoundError(IRRQueryError):
    proto_msg = "Key not found. (D)"


class KeyNotUniqueError(IRRQueryError):
    proto_msg = "There are multiple copies of the key in one database. (E)"


class OtherError(IRRQueryError):
    proto_msg = "Some other error, see the <optional message> for details."



from setuptools import setup, find_packages
from codecs import open

version = open('packaging/VERSION').read().strip()
requirements = open('packaging/requirements.txt').read().split("\n")
test_requirements = open('packaging/requirements-test.txt').read().split("\n")
with open('README.rst', encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='rptk',
    version=version,
    author='Workonline Communications',
    author_email='communications@workonkonline.co.za',
    description='Python tools for prefix filter list management operations',
    long_description=long_description,
    license='LICENSE',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Internet',
    ],
    packages=find_packages(
        include=[
            'rptk',
            'rptk.*'
        ],
        exclude=[]
    ),
    entry_points={
        'console_scripts': ['rptk=rptk.command_line:main']
    },
    include_package_data=True,

    url='https://github.com/wolcomm/rptk',
    download_url='https://github.com/wolcomm/rptk/%s' % version,

    install_requires=requirements,
    tests_require=test_requirements,
    test_suite='nose.collector'
)



import logging
from time import sleep
from rptk import RptkAPI
try:
    from daemonize import Daemonize
    can_daemonize = True
except ImportError:
    can_daemonize = False


def init():
    if can_daemonize:
        pid_file = __name__ + ".pid"
        daemon = Daemonize(app=__name__, action=main, pid=pid_file)
        daemon.start()
    else:
        main()


def main():
    # start logging
    log_file = __name__ + ".log"
    lh = logging.FileHandler(log_file)
    lf = logging.Formatter(fmt="%(asctime)s %(name)s: %(levelname)s %(message)s")
    lh.setFormatter(lf)
    logging.getLogger().addHandler(lh)
    # setup rptk api
    rptk = RptkAPI(format="ios")
    # begin loop
    while True:
        print rptk.query(obj="AS37271")
        sleep(3)
    return

if __name__ == "__main__":
    init()



from rptk.modules.format import BaseFormat


class TestFormat(BaseFormat):
    def format(self, result=None, name=None):
        self.log_method_enter(method=self.current_method)
        name = super(TestFormat, self).format(result=result, name=name)
        output = {
            name: result
        }
        self.log_method_exit(method=self.current_method)
        return output

    def validate(self, output=None):
        self.log_method_enter(method=self.current_method)
        if not isinstance(output, dict):
            self.raise_type_error(arg=output, cls=dict)
        self.log_method_exit(method=self.current_method)
        return True



import importlib
from rptk.base import BaseObject


class ClassLoader(BaseObject):
    def __init__(self, items=None):
        super(ClassLoader, self).__init__()
        self.log_init()
        if not isinstance(items, list):
            raise TypeError("%s not of type %s" % (items, list))
        self._classes = dict()
        count = 0
        self.log.debug(msg="trying to load classes")
        for item in items:
            name = item[0]
            mod_path, cls_path = item[1].rsplit(".", 1)
            self.log.debug(msg="loading class %s" % cls_path)
            try:
                cls = getattr(importlib.import_module(mod_path), cls_path)
                self._classes.update({name: cls})
                count += 1
            except Exception as e:
                self.log.warning(msg=e.message)
        self.log.debug(msg="loaded %s classes" % count)
        self.log_init_done()

    def get_class(self, name=None):
        return self._classes[name]

    @property
    def class_names(self):
        return [name for name in self._classes]

    @property
    def classes(self):
        return [self.get_class(name) for name in self.class_names]

    @property
    def log(self):
        return self._log



import json
import subprocess
from whichcraft import which
from rptk.modules.query import BaseQuery


class Bgpq3Query(BaseQuery):
    posix_only = True

    def query(self, obj=None):
        obj = super(Bgpq3Query, self).query(obj=obj)
        result = dict()
        if not self.path:
            msg = "couldn't determine bgpq3 executable path"
            self.log.error(msg=msg)
            raise RuntimeError(msg)
        cmds = {
            'ipv4': [self.path, "-h", self.target, "-l", "ipv4", "-4Aj", obj],
            'ipv6': [self.path, "-h", self.target, "-l", "ipv6", "-6Aj", obj]
        }
        for key in cmds:
            self.log.debug(msg="running %s" % ' '.join(cmds[key]))
            result.update(json.loads(subprocess.check_output(cmds[key])))
        self.log_method_exit(method=self.current_method)
        return result

    @property
    def path(self):
        self.log.debug(msg="determining bgpq3 executable path")
        try:
            return self.opts["bgpq3_path"]
        except KeyError:
            self.log.debug(msg="no configured path, using system default")
            return which("bgpq3")



import logging
import inspect


class BaseObject(object):
    def __init__(self):
        self._log = logging.getLogger(self.__module__)

    def __repr__(self):
        return "%s() object" % self.cls_name

    def __enter__(self):
        self.log_ready_start()
        self.log_ready_done()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.log_exit_start()
        self.log_exit_done()

    @property
    def opts(self):
        return getattr(self, "_opts", None)

    @property
    def log(self):
        return self._log

    @property
    def cls_name(self):
        return self.__class__.__name__

    @property
    def current_method(self):
        return inspect.currentframe().f_back.f_code.co_name

    def log_init(self):
        self.log.debug(msg="initialising %s instance" % self.cls_name)

    def log_init_done(self):
        caller = inspect.currentframe().f_back.f_back.f_code.co_name
        if caller == '__init__':
            self.log.debug(msg="still initialising %s instance" % self.cls_name)
        else:
            self.log.debug(msg="%s instance initialised" % self.cls_name)

    def log_method_enter(self, method=None):
        self.log.debug(msg="entering method %s.%s" % (self.cls_name, method))

    def log_method_exit(self, method=None):
        self.log.debug(msg="leaving method %s.%s" % (self.cls_name, method))

    def log_ready_start(self):
        self.log.debug(msg="preparing %s for use" % self)

    def log_ready_done(self):
        self.log.debug(msg="%s ready for use" % self)

    def log_exit_start(self):
        self.log.debug(msg="cleaning up %s" % self)

    def log_exit_done(self):
        self.log.debug(msg="finished cleaning up %s" % self)

    def raise_type_error(self, arg=None, cls=None):
        msg = "argument %s (%s) not of type %s" % (
            arg.__name__, arg, cls
            )
        self.log.error(msg=msg)
        raise TypeError(msg)

    def raise_runtime_error(self, msg=None):
        self.log.error(msg=msg)
        raise RuntimeError(msg)
